#!/usr/bin/env python
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This script generates a data source file for the C++ compressor.

Usage:
>>> ./gen_data.py /path/to/font.otf /path/to/output
"""

import array
import subprocess
import sys
from fontTools.ttLib import TTFont

from compreffor.cxxCompressor import write_data

if __name__ == '__main__':
    if len(sys.argv) < 3:
        print "missing arguments"
    else:
        f = TTFont(sys.argv[1])
        td = f['CFF '].cff.topDictIndex[0]
        with open(sys.argv[2], 'w') as out_f:
            data = write_data(td)
            out_f.write(data)
